@@ -1,18 +1,21 @@
-from django.conf import settings
 from django.contrib import admin
+from django.contrib.admin.filterspecs import AllValuesFilterSpec, FilterSpec
 from django.contrib.admin.util import unquote
 from django.contrib.admin.views.main import ChangeList, Paginator
+from django.core.cache import cache
 from django.forms.util import flatatt
 from django.http import HttpResponse
+from django.utils.encoding import force_unicode, smart_unicode
+from django.utils.html import escape
 from django.utils.safestring import mark_safe
-from django.utils.encoding import force_unicode
+from django.utils.translation import ugettext_lazy as _
 from django import forms
 
 from djangodblog.models import ErrorBatch, Error
 from djangodblog.helpers import ImprovedExceptionReporter
+from djangodblog.settings import *
 from djangodblog.utils import JSONDictFormField
 
-import cgi
 import base64
 import re
 import logging
@@ -24,41 +27,7 @@ except ImportError:
 
 logger = logging.getLogger('dblog')
 
-class PreformattedText(forms.Textarea):
-    input_type = 'textarea'
-    
-    def render(self, name, value, attrs=None):
-        if value is None: value = ''
-        if value != '':
-            # Only add the 'value' attribute if a value is non-empty.
-            value = force_unicode(value)
-        return mark_safe(u'<pre style="clear:left;display:block;padding-top:5px;white-space: pre-wrap;white-space: -moz-pre-wrap;white-space: -pre-wrap;white-space: -o-pre-wrap;word-wrap: break-word;">%s</pre>' % (value,))
-
-class Link(forms.TextInput):
-    input_type = 'a'
-
-    def render(self, name, value, attrs=None):
-        if value is None: value = ''
-        if value != '':
-            # Only add the 'value' attribute if a value is non-empty.
-            value = force_unicode(value)
-        return mark_safe(u'<a href="%s">%s</a>' % (value, cgi.escape(value)))
-
-class ErrorBatchAdminForm(forms.ModelForm):
-    traceback = forms.CharField(widget=PreformattedText())
-    url = forms.CharField(widget=Link())
-    
-    class Meta:
-        fields = ('url', 'logger', 'server_name', 'class_name', 'level', 'message', 'times_seen', 'first_seen', 'last_seen', 'traceback')
-        model = ErrorBatch
-
-class ErrorAdminForm(forms.ModelForm):
-    traceback = forms.CharField(widget=PreformattedText())
-    url = forms.CharField(widget=Link())
-    
-    class Meta:
-        fields = ('url', 'logger', 'server_name', 'class_name', 'level', 'message', 'datetime', 'traceback')
-        model = ErrorBatch
+# Custom admin and pagination clauses for efficiency
 
 class EfficientPaginator(Paginator):
     def _get_count(self):
@@ -93,23 +62,87 @@ class EfficientModelAdmin(admin.ModelAdmin):
     def get_changelist(self, request, **kwargs):
         return EfficientChangeList
 
+class CachedAllValuesFilterSpec(AllValuesFilterSpec):
+    def __init__(self, f, request, params, model, model_admin):
+        super(AllValuesFilterSpec, self).__init__(f, request, params, model, model_admin)
+        self.lookup_val = request.GET.get(f.name, None)
+        cache_key = 'admin_filters_%s_%s' % (model.__name__, f.name)
+        self.lookup_choices = cache.get(cache_key)
+        if self.lookup_choices is None:
+            self.lookup_choices = list(model_admin.queryset(request).distinct().order_by(f.name).values_list(f.name, flat=True))
+            cache.set(cache_key, self.lookup_choices, 60*5)
+
+    def title(self):
+        return self.field.verbose_name
+
+    def choices(self, cl):
+        yield {'selected': self.lookup_val is None,
+               'query_string': cl.get_query_string({}, [self.field.name]),
+               'display': _('All')}
+        for val in self.lookup_choices:
+            val = smart_unicode(val)
+            yield {'selected': self.lookup_val == val,
+                   'query_string': cl.get_query_string({self.field.name: val}),
+                   'display': val}
+FilterSpec.filter_specs.insert(-1, (lambda f: f.model._meta.app_label == 'djangodblog', CachedAllValuesFilterSpec))
+
 UNDEFINED = object()
 
 class FakeRequest(object):
     def build_absolute_uri(self): return self.url
+    
+# Custom forms/fields for the admin
+
+class PreformattedText(forms.Textarea):
+    input_type = 'textarea'
+    
+    def render(self, name, value, attrs=None):
+        if value is None: value = ''
+        if value != '':
+            # Only add the 'value' attribute if a value is non-empty.
+            value = force_unicode(value)
+        return mark_safe(u'<pre style="clear:left;display:block;padding-top:5px;white-space: pre-wrap;white-space: -moz-pre-wrap;white-space: -pre-wrap;white-space: -o-pre-wrap;word-wrap: break-word;">%s</pre>' % (escape(value),))
+
+class Link(forms.TextInput):
+    input_type = 'a'
+
+    def render(self, name, value, attrs=None):
+        if value is None: value = ''
+        if value != '':
+            # Only add the 'value' attribute if a value is non-empty.
+            value = force_unicode(value)
+        return mark_safe(u'<a href="%s">%s</a>' % (value, escape(value)))
+
+class ErrorBatchAdminForm(forms.ModelForm):
+    traceback = forms.CharField(widget=PreformattedText())
+    url = forms.CharField(widget=Link())
+    
+    class Meta:
+        fields = ('url', 'logger', 'server_name', 'class_name', 'level', 'message', 'times_seen', 'first_seen', 'last_seen', 'traceback')
+        model = ErrorBatch
+
+class ErrorAdminForm(forms.ModelForm):
+    traceback = forms.CharField(widget=PreformattedText())
+    url = forms.CharField(widget=Link())
+    
+    class Meta:
+        fields = ('url', 'logger', 'server_name', 'class_name', 'level', 'message', 'datetime', 'traceback')
+        model = ErrorBatch
+
+# Actual admin modules
 
 class ErrorBatchAdmin(EfficientModelAdmin):
     form            = ErrorBatchAdminForm
     list_display    = ('shortened_url', 'logger', 'level', 'server_name', 'times_seen', 'last_seen')
     list_display_links = ('shortened_url',)
-    list_filter     = ('status', 'level', 'last_seen')
+    list_filter     = ('status', 'server_name', 'logger', 'level', 'last_seen')
     ordering        = ('-last_seen',)
     actions         = ('resolve_errorbatch',)
     search_fields   = ('url', 'class_name', 'message', 'traceback', 'server_name')
-    readonly_fields = ('logger', 'server_name', 'class_name', 'level', 'message', 'times_seen', 'first_seen', 'last_seen')
+    readonly_fields = ('class_name', 'message', 'times_seen', 'first_seen')
     fieldsets       = (
         (None, {
-            'fields': ('url', 'logger', 'server_name', 'class_name', 'level', 'message', 'times_seen', 'first_seen', 'last_seen', 'traceback')
+            'fields': ('class_name', 'message', 'times_seen', 'first_seen', 'traceback')
         }),
     )
     
@@ -124,17 +157,26 @@ class ErrorBatchAdmin(EfficientModelAdmin):
         
     resolve_errorbatch.short_description = 'Resolve selected error summaries'
 
+    def change_view(self, request, object_id, extra_context={}):
+        obj = self.get_object(request, unquote(object_id))
+        recent_errors = Error.objects.filter(checksum=obj.checksum).order_by('-datetime')[0:5]
+        extra_context.update({
+            'instance': obj,
+            'recent_errors': recent_errors,
+        })
+        return super(ErrorBatchAdmin, self).change_view(request, object_id, extra_context)
+
 class ErrorAdmin(EfficientModelAdmin):
     form            = ErrorAdminForm
     list_display    = ('shortened_url', 'logger', 'level', 'server_name', 'datetime')
     list_display_links = ('shortened_url',)
-    list_filter     = ('level', 'datetime')
+    list_filter     = ('server_name', 'logger', 'level', 'datetime')
     ordering        = ('-id',)
     search_fields   = ('url', 'class_name', 'message', 'traceback', 'server_name')
-    readonly_fields = ('logger', 'server_name', 'class_name', 'level', 'message', 'datetime')
+    readonly_fields = ('class_name', 'message')
     fieldsets       = (
         (None, {
-            'fields': ('url', 'logger', 'server_name', 'class_name', 'level', 'message', 'datetime', 'traceback')
+            'fields': ('class_name', 'message', 'traceback')
         }),
     )
     
@@ -143,8 +185,10 @@ class ErrorAdmin(EfficientModelAdmin):
     
     def change_view(self, request, object_id, extra_context={}):
         obj = self.get_object(request, unquote(object_id))
-        has_traceback = getattr(settings, 'DBLOG_ENHANCED_TRACEBACKS', True) and bool('exc' in obj.data) and 'raw' not in request.GET
-        if has_traceback:
+        has_traceback = ENHANCED_TRACEBACKS \
+                and 'exc' in obj.data and 'META' in obj.data
+        show_traceback = has_traceback and 'raw' not in request.GET
+        if show_traceback:
             try:
                 extra_context.update(self.get_traceback_context(request, obj))
             except Exception:
@@ -153,6 +197,7 @@ class ErrorAdmin(EfficientModelAdmin):
                 has_traceback = False
         extra_context.update({
             'has_traceback': has_traceback,
+            'show_traceback': show_traceback,
             'instance': obj,
         })
         return super(ErrorAdmin, self).change_view(request, object_id, extra_context)
@@ -180,7 +225,7 @@ class ErrorAdmin(EfficientModelAdmin):
                 exc_type = type(obj.class_name, (Exception,), {})
         exc_value = exc_type(obj.message)
         exc_value.args = args
-    
+        
         fake_request = FakeRequest()
         fake_request.META = obj.data['META']
         fake_request.GET = obj.data['GET']
