@@ -14,18 +14,17 @@ try:
 except ImportError:
     import pickle
 
-from django.conf import settings
+from django.core.cache import cache
 from django.db import models
-from django.conf import settings
 from django.db.models import sql
-from django.utils.hashcompat import md5_constructor
-from django.utils.encoding import smart_unicode
 from django.db.models.query import QuerySet
+from django.utils.encoding import smart_unicode
 from django.views.debug import ExceptionReporter
 
-DBLOG_DATABASE_USING = getattr(settings, 'DBLOG_DATABASE_USING', None)
+from djangodblog import settings
+from djangodblog.helpers import construct_checksum
 
-assert not DBLOG_DATABASE_USING or django.VERSION >= (1, 2), 'The `DBLOG_DATABASE_USING` setting requires Django >= 1.2'
+assert not settings.DATABASE_USING or django.VERSION >= (1, 2), 'The `DBLOG_DATABASE_USING` setting requires Django >= 1.2'
 
 logger = logging.getLogger('dblog')
 
@@ -34,33 +33,42 @@ class DBLogManager(models.Manager):
 
     def get_query_set(self):
         qs = super(DBLogManager, self).get_query_set()
-        if DBLOG_DATABASE_USING:
-            qs = qs.using(DBLOG_DATABASE_USING)
+        if settings.DATABASE_USING:
+            qs = qs.using(settings.DATABASE_USING)
         return qs
 
     def _create(self, **defaults):
         from models import Error, ErrorBatch
         
+        URL_MAX_LENGTH = Error._meta.get_field_by_name('url')[0].max_length
+        
         server_name = socket.gethostname()
         class_name  = defaults.pop('class_name', None)
-        checksum    = md5_constructor(str(defaults.get('level', logging.FATAL)))
-        checksum.update(class_name or '')
-        checksum.update(defaults.get('traceback') or defaults['message'])
-        checksum    = checksum.hexdigest()
+        
+        data = defaults.pop('data', {}) or {}
+        if defaults.get('url'):
+            data['url'] = defaults['url']
+            defaults['url'] = defaults['url'][:URL_MAX_LENGTH]
 
-        data = defaults.pop('data', {})
+        instance = Error(
+            class_name=class_name,
+            server_name=server_name,
+            data=data,
+            **defaults
+        )
+        instance.checksum = construct_checksum(instance)
+        
+        cache_key = 'djangodblog:%s:%s' % (instance.class_name, instance.checksum)
+        added = cache.add(cache_key, 1, settings.THRASHING_TIMEOUT)
+        if not added and cache.incr(cache_key) > settings.THRASHING_LIMIT:
+            return
 
         try:
-            instance = Error.objects.create(
-                class_name=class_name,
-                server_name=server_name,
-                data=data,
-                **defaults
-            )
+            instance.save()
             batch, created = ErrorBatch.objects.get_or_create(
                 class_name = class_name,
                 server_name = server_name,
-                checksum = checksum,
+                checksum = instance.checksum,
                 defaults = defaults
             )
             if not created:
@@ -70,7 +78,10 @@ class DBLogManager(models.Manager):
                     last_seen=datetime.datetime.now(),
                 )
         except Exception, exc:
-            logger.exception(exc)
+            try:
+                logger.exception(exc)
+            except Exception, exc:
+                warnings.warn(u'Unable to process log entry: %s' % (exc,))
         else:
             return instance
     
@@ -78,7 +89,7 @@ class DBLogManager(models.Manager):
         """
         Creates an error log for a `logging` module `record` instance.
         """
-        for k in ('url',):
+        for k in ('url', 'data'):
             if k not in kwargs:
                 kwargs[k] = record.__dict__.get(k)
         kwargs.update({
@@ -125,22 +136,22 @@ class DBLogManager(models.Manager):
             elif isinstance(f, (list, tuple)):
                 f = [to_unicode(f) for f in f]
             else:
-                f = unicode(f)
+                f = smart_unicode(f)
             return f
 
         reporter = ExceptionReporter(None, exc_type, exc_value, traceback)
         frames = reporter.get_traceback_frames()
 
-        data = kwargs.pop('data', {})
+        data = kwargs.pop('data', {}) or {}
         data['exc'] = base64.b64encode(pickle.dumps(map(to_unicode, [exc_type.__class__.__module__, exc_value.args, frames])).encode('zlib'))
 
         tb_message = '\n'.join(traceback_mod.format_exception(exc_type, exc_value, traceback))
 
-        kwargs.setdefault('message', smart_unicode(exc_value))
+        kwargs.setdefault('message', to_unicode(exc_value))
 
         return self._create(
             class_name=exc_type.__name__,
             traceback=tb_message,
             data=data,
             **kwargs
-        )
\ No newline at end of file
+        )
